# -*- coding:utf-8 -*-
from __future__ import unicode_literals

from time import sleep
from unittest import skipUnless

import pytest
from django.test import SimpleTestCase, TestCase
from django.utils import six

from django_mysql.utils import (
    PTFingerprintThread, WeightedAverageRate, format_duration, have_program,
    index_name, pt_fingerprint
)
from testapp.models import Author, AuthorMultiIndex


class WeightedAverageRateTests(SimpleTestCase):

    def test_constant(self):
        # If we keep achieving a rate of 100 rows in 0.5 seconds, it should
        # recommend that we keep there
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 0.5) == 100
        assert rate.update(100, 0.5) == 100
        assert rate.update(100, 0.5) == 100

    def test_slow(self):
        # If we keep achieving a rate of 100 rows in 1 seconds, it should
        # recommend that we move to 50
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 1.0) == 50
        assert rate.update(100, 1.0) == 50
        assert rate.update(100, 1.0) == 50

    def test_fast(self):
        # If we keep achieving a rate of 100 rows in 0.25 seconds, it should
        # recommend that we move to 200
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 0.25) == 200
        assert rate.update(100, 0.25) == 200
        assert rate.update(100, 0.25) == 200

    def test_good_guess(self):
        # If we are first slow then hit the target at 50, we should be good
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 1.0) == 50
        assert rate.update(50, 0.5) == 50
        assert rate.update(50, 0.5) == 50
        assert rate.update(50, 0.5) == 50

    def test_zero_division(self):
        rate = WeightedAverageRate(0.5)
        assert rate.update(1, 0.0) == 500


class FormatDurationTests(SimpleTestCase):

    def test_seconds(self):
        assert format_duration(0) == '0s'
        assert format_duration(1) == '1s'
        assert format_duration(30) == '30s'
        assert format_duration(59) == '59s'

    def test_minutes(self):
        assert format_duration(60) == '1m0s'
        assert format_duration(61) == '1m1s'
        assert format_duration(120) == '2m0s'
        assert format_duration(3599) == '59m59s'

    def test_hours(self):
        assert format_duration(3600) == '1h0m0s'
        assert format_duration(3601) == '1h0m1s'


@skipUnless(have_program('pt-fingerprint'),
            "pt-fingerprint must be installed")
class PTFingerprintTests(SimpleTestCase):

    def test_basic(self):
        assert pt_fingerprint('SELECT 5') == 'select ?'
        assert pt_fingerprint('SELECT 5;') == 'select ?'

    def test_long(self):
        query = """
            SELECT
                CONCAT(customer.last_name, ', ', customer.first_name)
                    AS customer,
                address.phone,
                film.title
            FROM rental
                INNER JOIN customer
                    ON rental.customer_id = customer.customer_id
                INNER JOIN address
                    ON customer.address_id = address.address_id
                INNER JOIN inventory
                    ON rental.inventory_id = inventory.inventory_id
                INNER JOIN film
                    ON inventory.film_id = film.film_id
            WHERE
                rental.return_date IS NULL AND
                rental_date + INTERVAL film.rental_duration DAY <
                    CURRENT_DATE()
            LIMIT 5"""
        assert (
            pt_fingerprint(query) ==
            "select concat(customer.last_name, ?, customer.first_name) as "
            "customer, address.phone, film.title from rental inner join "
            "customer on rental.customer_id = customer.customer_id inner join "
            "address on customer.address_id = address.address_id inner join "
            "inventory on rental.inventory_id = inventory.inventory_id inner "
            "join film on inventory.film_id = film.film_id where "
            "rental.return_date is ? and rental_date ? interval "
            "film.rental_duration day < current_date() limit ?"
        )

    def test_the_thread_shuts_on_time_out(self):
        PTFingerprintThread.PROCESS_LIFETIME = 0.1
        pt_fingerprint("select 123")
        sleep(0.2)
        assert PTFingerprintThread.the_thread is None
        PTFingerprintThread.PROCESS_LIFETIME = 60


class IndexNameTests(TestCase):
    def test_requires_field_names(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author)
        assert (
            "At least one field name required" in
            six.text_type(excinfo.value)
        )

    def test_requires_real_field_names(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author, 'nonexistent')
        assert (
            "Fields do not exist: nonexistent" in
            six.text_type(excinfo.value)
        )

    def test_invalid_kwarg(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author, 'name', nonexistent_kwarg=True)
        assert (
            "The only supported keyword argument is 'using'" in
            six.text_type(excinfo.value)
        )

    def test_primary_key(self):
        assert index_name(Author, 'id') == 'PRIMARY'

    def test_primary_key_using_other(self):
        assert index_name(Author, 'id', using='other') == 'PRIMARY'

    def test_secondary_single_field(self):
        name = index_name(Author, 'name')
        assert name.startswith('testapp_author_')

    def test_index_does_not_exist(self):
        with pytest.raises(KeyError) as excinfo:
            index_name(Author, 'bio')
        assert "There is no index on (bio)" in six.text_type(excinfo.value)

    def test_secondary_multiple_fields(self):
        name = index_name(AuthorMultiIndex, 'name', 'country')
        assert name.startswith('testapp_authormultiindex')

    def test_secondary_multiple_fields_non_existent_reversed_existent(self):
        # Checks that order is preserved
        with pytest.raises(KeyError):
            index_name(AuthorMultiIndex, 'country', 'name')

    def test_secondary_multiple_fields_non_existent(self):
        with pytest.raises(KeyError):
            index_name(AuthorMultiIndex, 'country', 'id')



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import django
from django.db import connection


def pytest_report_header(config):
    dot_version = '.'.join(str(x) for x in django.VERSION)
    header = "Django version: " + dot_version

    if hasattr(connection, '_nodb_connection'):
        with connection._nodb_connection.cursor() as cursor:
            cursor.execute("SELECT VERSION()")
            version = cursor.fetchone()[0]
        header += "\nMySQL version: {}".format(version)

    return header



# -*- encoding:utf-8 -*-
from __future__ import absolute_import, division, unicode_literals

import pytest
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.db import connection
from django.db.utils import DataError
from django.test import TestCase, TransactionTestCase, override_settings

from django_mysql.models import EnumField
from testapp.models import EnumModel, NullableEnumModel


class TestEnumField(TestCase):
    def test_empty_choices(self):
        with pytest.raises(ValueError):
            EnumField(choices=[])

    def test_no_choices(self):
        with pytest.raises(ValueError):
            EnumField()

    def test_invalid_choices(self):
        with pytest.raises(TypeError) as exc_info:
            EnumField(choices=['red', 10])
        assert 'Invalid choice "10"' in str(exc_info.value)

    def test_invalid_max_length(self):
        with pytest.raises(TypeError) as exc_info:
            EnumField(choices=['red'], max_length=100)
        assert '"max_length" is not a valid argument' in str(exc_info.value)

    def test_correct(self):
        s = EnumModel.objects.create(field='red')
        assert s.field == 'red'
        s = EnumModel.objects.get(id=s.id)
        assert s.field == 'red'

    def test_invalid_value(self):
        with pytest.raises(DataError):
            EnumModel.objects.create(field='elephant')

    def test_empty(self):
        with pytest.raises(DataError):
            EnumModel.objects.create()

    def test_null_storage(self):
        s = NullableEnumModel.objects.create()
        assert s.field is None
        s = NullableEnumModel.objects.get(id=s.id)
        assert s.field is None

    def test_isnull_lookup(self):
        NullableEnumModel.objects.create(field='goat')
        s = NullableEnumModel.objects.create()
        actual = NullableEnumModel.objects.filter(field__isnull=True)
        expected = [s]

        assert list(actual) == expected

    def test_basic_lookup(self):
        s1 = EnumModel.objects.create(field='green')
        EnumModel.objects.create(field='red')
        s2 = EnumModel.objects.create(field='green')

        actual = EnumModel.objects.filter(field='green')
        expected = [s1, s2]

        assert list(actual) == expected

    def test_in_lookup(self):
        s1 = EnumModel.objects.create(field='green')
        EnumModel.objects.create(field='red')
        s2 = EnumModel.objects.create(field='green')

        actual = EnumModel.objects.filter(field__in=['green'])
        expected = [s1, s2]

        assert list(actual) == expected

    def test_icontains_lookup(self):
        s1 = EnumModel.objects.create(field='coralBlue')
        s2 = EnumModel.objects.create(field='blue')
        EnumModel.objects.create(field='green')

        actual = EnumModel.objects.filter(field__icontains='Blue')
        expected = [s1, s2]

        assert list(actual) == expected

    def test_contains_lookup(self):
        s1 = EnumModel.objects.create(field='coralblue')
        s2 = EnumModel.objects.create(field='blue')
        EnumModel.objects.create(field='green')

        actual = EnumModel.objects.filter(field__contains='blue')
        expected = [s1, s2]

        assert list(actual) == expected


class TestCheck(TestCase):

    def test_db_not_mariadb(self):
        errors = EnumModel.check()
        assert errors == []


class TestDeconstruct(TestCase):

    def test_deconstruct(self):
        field = EnumField(choices=['a', 'b'])
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.EnumField'
        assert 'max_length' not in kwargs
        EnumField(*args, **kwargs)


class TestMigrations(TransactionTestCase):

    @override_settings(MIGRATION_MODULES={
        "testapp": "testapp.enum_default_migrations",
    })
    def test_adding_field_with_default(self):
        table_name = 'testapp_enumdefaultmodel'
        table_names = connection.introspection.table_names
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)

        call_command('migrate', 'testapp',
                     verbosity=0, skip_checks=True)
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)

        call_command('migrate', 'testapp', 'zero',
                     verbosity=0, skip_checks=True)
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)


class TestFormfield(TestCase):
    def test_formfield(self):
        model_field = EnumField(choices=['this', 'that'])
        form_field = model_field.formfield()

        assert form_field.clean('this') == 'this'
        assert form_field.clean('that') == 'that'

        with pytest.raises(ValidationError):
            form_field.clean('')

        with pytest.raises(ValidationError):
            form_field.clean('invalid')



# -*- coding:utf-8 -*-
from __future__ import absolute_import, unicode_literals

import collections
import json

import django
from django.core import checks
from django.db import connections
from django.db.models import Field, IntegerField, Transform
from django.utils import six

from django_mysql import forms
from django_mysql.compat import field_class
from django_mysql.models.lookups import (
    JSONContainedBy, JSONContains, JSONHasAnyKeys, JSONHasKey, JSONHasKeys
)
from django_mysql.utils import collapse_spaces

__all__ = ('JSONField',)


class JSONField(field_class(Field)):
    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = dict
        super(JSONField, self).__init__(*args, **kwargs)

    def check(self, **kwargs):
        errors = super(JSONField, self).check(**kwargs)
        errors.extend(self._check_django_version())
        errors.extend(self._check_default())

        if django.VERSION[:2] >= (1, 8):
            # This check connects to the DB, which only works on Django 1.8+
            errors.extend(self._check_mysql_version())

        return errors

    def _check_django_version(self):
        errors = []
        if django.VERSION[:2] < (1, 8):
            errors.append(
                checks.Error(
                    "Django 1.8+ is required to use JSONField",
                    obj=self,
                    id='django_mysql.E015',
                )
            )
        return errors

    def _check_default(self):
        errors = []
        if isinstance(self.default, (list, dict)):
            errors.append(
                checks.Error(
                    'Do not use mutable defaults for JSONField',
                    hint=collapse_spaces('''
                        Mutable defaults get shared between all instances of
                        the field, which probably isn't what you want. You
                        should replace your default with a callable, e.g.
                        replace default={{}} with default=dict.

                        The default you passed was '{}'.
                    '''.format(self.default)),
                    obj=self,
                    id='django_mysql.E017',
                )
            )
        return errors

    def _check_mysql_version(self):
        errors = []

        any_conn_works = False
        conn_names = ['default'] + list(set(connections) - {'default'})
        for db in conn_names:
            if (
                not connections[db].is_mariadb and
                connections[db].mysql_version >= (5, 7)
            ):
                any_conn_works = True

        if not any_conn_works:
            errors.append(
                checks.Error(
                    "MySQL 5.7+ is required to use JSONField",
                    hint=None,
                    obj=self,
                    id='django_mysql.E016'
                )
            )
        return errors

    def deconstruct(self):
        name, path, args, kwargs = super(JSONField, self).deconstruct()
        path = 'django_mysql.models.%s' % self.__class__.__name__
        return name, path, args, kwargs

    def db_type(self, connection):
        return 'json'

    def get_transform(self, name):
        transform = super(JSONField, self).get_transform(name)
        if transform:
            return transform  # pragma: no cover
        return KeyTransformFactory(name)

    def from_db_value(self, value, expression, connection, context):
        # Similar to to_python, for Django 1.8+
        if isinstance(value, six.string_types):
            return json.loads(value)
        return value

    def get_prep_value(self, value):
        if value is not None and not isinstance(value, six.string_types):
            # For some reason this value gets string quoted in Django's SQL
            # compiler...
            return json.dumps(value)
        return value

    def get_prep_lookup(self, lookup_type, value):
        if (
            not hasattr(value, '_prepare') and
            lookup_type in ('exact', 'gt', 'gte', 'lt', 'lte', 'contains') and
            value is not None
        ):
            return JSONValue(value)
        elif lookup_type == 'has_key':
            if not isinstance(value, six.text_type):
                raise ValueError(
                    "JSONField's 'has_key' lookup only works with {} values"
                    .format(six.text_type.__name__)
                )
            return value
        elif lookup_type in ('has_keys', 'has_any_keys'):
            if not isinstance(value, collections.Sequence):
                raise ValueError(
                    "JSONField's '{}' lookup only works with Sequences"
                    .format(lookup_type)
                )
            return value

        return super(JSONField, self).get_prep_lookup(lookup_type, value)

    def get_lookup(self, lookup_name):
        # Have to 'unregister' some incompatible lookups
        if lookup_name in {
            'range', 'in', 'iexact', 'icontains', 'startswith',
            'istartswith', 'endswith', 'iendswith', 'search', 'regex', 'iregex'
        }:
            raise NotImplementedError(
                "Lookup '{}' doesn't work with JSONField".format(lookup_name)
            )
        return super(JSONField, self).get_lookup(lookup_name)

    def formfield(self, **kwargs):
        defaults = {'form_class': forms.JSONField}
        defaults.update(kwargs)
        return super(JSONField, self).formfield(**defaults)


class JSONValue(object):
    def __init__(self, value):
        self.json_string = json.dumps(value)

    def as_sql(self, *args, **kwargs):
        return 'CAST(%s AS JSON)', [self.json_string]


class JSONLength(Transform):
    lookup_name = 'length'

    output_field = IntegerField()

    if django.VERSION[:2] < (1, 9):
        def as_sql(self, compiler, connection):
            lhs, params = compiler.compile(self.lhs)
            return 'JSON_LENGTH({})'.format(lhs), params
    else:
        function = 'JSON_LENGTH'

JSONField.register_lookup(JSONContains)
JSONField.register_lookup(JSONContainedBy)
JSONField.register_lookup(JSONHasAnyKeys)
JSONField.register_lookup(JSONHasKey)
JSONField.register_lookup(JSONHasKeys)
JSONField.register_lookup(JSONLength)


class KeyTransform(Transform):

    def __init__(self, key_name, *args, **kwargs):
        super(KeyTransform, self).__init__(*args, **kwargs)
        self.key_name = key_name

    def as_sql(self, compiler, connection):
        key_transforms = [self.key_name]
        previous = self.lhs
        while isinstance(previous, KeyTransform):
            key_transforms.insert(0, previous.key_name)
            previous = previous.lhs

        lhs, params = compiler.compile(previous)

        json_path = self.compile_json_path(key_transforms)

        return 'JSON_EXTRACT({}, %s)'.format(lhs), params + [json_path]

    def compile_json_path(self, key_transforms):
        path = ['$']
        for key_transform in key_transforms:
            try:
                num = int(key_transform)
                path.append('[{}]'.format(num))
            except ValueError:  # non-integer
                path.append('.')
                path.append(key_transform)
        return ''.join(path)


class KeyTransformFactory(object):

    def __init__(self, key_name):
        self.key_name = key_name

    def __call__(self, *args, **kwargs):
        return KeyTransform(self.key_name, *args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models

from django_mysql.models import ListCharField


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='intlistdefaultmodel',
            name='field_2',
            field=ListCharField(models.IntegerField(),
                                default=lambda: [],
                                size=None,
                                max_length=32),
        ),
        migrations.AddField(
            model_name='intlistdefaultmodel',
            name='field_3',
            field=ListCharField(models.IntegerField(),
                                default=lambda: [1, 5],
                                size=None,
                                max_length=32),
        ),
    ]



import django
from django.db.models import Aggregate, CharField
from django.utils.functional import cached_property

# Major aggregate simplification from 1.7 to 1.8 - it's easier to implement
# each twice than try fudge a class that works before and after

if django.VERSION[:2] < (1, 8):

    from django.db.models.sql.aggregates import Aggregate as SQLAggregate

    class BitAnd(Aggregate):
        name = 'BitAnd'

        def add_to_query(self, query, alias, col, source, is_summary):
            query.aggregates[alias] = BitAndSQL(
                col,
                source=source,
                is_summary=is_summary,
                **self.extra
            )

    class BitAndSQL(SQLAggregate):
        sql_function = 'BIT_AND'
        is_ordinal = True  # is an integer

    class BitOr(Aggregate):
        name = 'BitOr'

        def add_to_query(self, query, alias, col, source, is_summary):
            query.aggregates[alias] = BitOrSQL(
                col,
                source=source,
                is_summary=is_summary,
                **self.extra
            )

    class BitOrSQL(SQLAggregate):
        sql_function = 'BIT_OR'
        is_ordinal = True  # is an integer

    class BitXor(Aggregate):
        name = 'BitXor'

        def add_to_query(self, query, alias, col, source, is_summary):
            query.aggregates[alias] = BitXorSQL(
                col,
                source=source,
                is_summary=is_summary,
                **self.extra
            )

    class BitXorSQL(SQLAggregate):
        sql_function = 'BIT_XOR'
        is_ordinal = True  # is an integer

    class GroupConcat(Aggregate):
        def add_to_query(self, query, alias, col, source, is_summary):
            query.aggregates[alias] = SQLGroupConcat(
                col,
                source=source,
                is_summary=is_summary,
                **self.extra
            )

    class SQLGroupConcat(SQLAggregate):

        def __init__(self, col, distinct=False, separator=None,
                     ordering=None, **extra):
            super(SQLGroupConcat, self).__init__(col, **extra)

            self.distinct = distinct
            self.separator = separator

            # This can/will be improved to SetTextField or ListTextField
            self.field = CharField()

            if ordering not in ('asc', 'desc', None):
                raise ValueError(
                    "'ordering' must be one of 'asc', 'desc', or None")
            self.ordering = ordering

        sql_function = 'GROUP_CONCAT'

        @cached_property
        def sql_template(self):
            # Constructing a template...

            template = ["%(function)s("]

            if self.distinct:
                template.append("DISTINCT ")

            template.append("%(field)s")

            if self.separator is not None:
                template.append(" SEPARATOR '{}'".format(self.separator))

            if self.ordering is not None:
                template.append(" ORDER BY %(field)s " + self.ordering.upper())

            template.append(")")

            return "".join(template)


else:

    class BitAnd(Aggregate):
        function = 'BIT_AND'
        name = 'bitand'

    class BitOr(Aggregate):
        function = 'BIT_OR'
        name = 'bitor'

    class BitXor(Aggregate):
        function = 'BIT_XOR'
        name = 'bitxor'

    class GroupConcat(Aggregate):
        function = 'GROUP_CONCAT'

        def __init__(self, expression, distinct=False, separator=None,
                     ordering=None, **extra):

            if 'output_field' not in extra:
                # This can/will be improved to SetTextField or ListTextField
                extra['output_field'] = CharField()

            super(GroupConcat, self).__init__(expression, **extra)

            self.distinct = distinct
            self.separator = separator

            if ordering not in ('asc', 'desc', None):
                raise ValueError(
                    "'ordering' must be one of 'asc', 'desc', or None")
            self.ordering = ordering

        def as_sql(self, compiler, connection, function=None, template=None):
            connection.ops.check_expression_support(self)
            sql = ["GROUP_CONCAT("]
            if self.distinct:
                sql.append("DISTINCT ")

            expr_parts = []
            params = []
            for arg in self.source_expressions:
                arg_sql, arg_params = compiler.compile(arg)
                expr_parts.append(arg_sql)
                params.extend(arg_params)
            expr_sql = self.arg_joiner.join(expr_parts)

            sql.append(expr_sql)

            if self.separator is not None:
                sql.append(" SEPARATOR '{}'".format(self.separator))

            if self.ordering is not None:
                sql.append(" ORDER BY ")
                sql.append(expr_sql)
                params.extend(params[:])
                sql.append(" ")
                sql.append(self.ordering.upper())

            sql.append(")")

            return "".join(sql), params



# -*- coding:utf-8 -*-
import json

import pytest
from django import forms
from django.core import exceptions, serializers
from django.db import models
from django.db.migrations.writer import MigrationWriter
from django.db.models import Q
from django.test import SimpleTestCase, TestCase
from django.utils import six

from django_mysql.forms import SimpleSetField
from django_mysql.models import SetTextField
from testapp.models import BigCharSetModel, BigIntSetModel, TemporaryModel


class TestSaveLoad(TestCase):

    def test_char_easy(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        s = BigCharSetModel.objects.create(field=big_set)
        assert s.field == big_set
        s = BigCharSetModel.objects.get(id=s.id)
        assert s.field == big_set

        letters = set('abcdefghi')
        bigger_set = big_set | letters
        s.field.update(letters)
        assert s.field == bigger_set
        s.save()
        s = BigCharSetModel.objects.get(id=s.id)
        assert s.field == bigger_set

    def test_char_string_direct(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        big_str = ','.join(big_set)
        s = BigCharSetModel.objects.create(field=big_str)
        s = BigCharSetModel.objects.get(id=s.id)
        assert s.field == big_set

    def test_is_a_set_immediately(self):
        s = BigCharSetModel()
        assert s.field == set()
        s.field.add("bold")
        s.field.add("brave")
        s.save()
        assert s.field == {"bold", "brave"}
        s = BigCharSetModel.objects.get(id=s.id)
        assert s.field == {"bold", "brave"}

    def test_empty(self):
        s = BigCharSetModel.objects.create()
        assert s.field == set()
        s = BigCharSetModel.objects.get(id=s.id)
        assert s.field == set()

    def test_char_cant_create_sets_with_commas(self):
        with pytest.raises(ValueError):
            BigCharSetModel.objects.create(field={"co,mma", "contained"})

    def test_char_cant_create_sets_with_empty_string(self):
        with pytest.raises(ValueError):
            BigCharSetModel.objects.create(field={""})

    def test_char_basic_lookup(self):
        mymodel = BigCharSetModel.objects.create()
        empty = BigCharSetModel.objects.filter(field="")

        assert empty.count() == 1
        assert empty[0] == mymodel

        mymodel.delete()

        assert empty.count() == 0

    def test_char_lookup_contains(self):
        self.check_char_lookup('contains')

    def test_char_lookup_icontains(self):
        self.check_char_lookup('icontains')

    def check_char_lookup(self, lookup):
        lname = 'field__' + lookup
        mymodel = BigCharSetModel.objects.create(field={"mouldy", "rotten"})

        mouldy = BigCharSetModel.objects.filter(**{lname: "mouldy"})
        assert mouldy.count() == 1
        assert mouldy[0] == mymodel

        rotten = BigCharSetModel.objects.filter(**{lname: "rotten"})
        assert rotten.count() == 1
        assert rotten[0] == mymodel

        clean = BigCharSetModel.objects.filter(**{lname: "clean"})
        assert clean.count() == 0

        with pytest.raises(ValueError):
            list(BigCharSetModel.objects.filter(**{lname: {"a", "b"}}))

        both = BigCharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) & Q(**{lname: "rotten"})
        )
        assert both.count() == 1
        assert both[0] == mymodel

        either = BigCharSetModel.objects.filter(
            Q(**{lname: "mouldy"}) | Q(**{lname: "clean"})
        )
        assert either.count() == 1

        not_clean = BigCharSetModel.objects.exclude(**{lname: "clean"})
        assert not_clean.count() == 1

        not_mouldy = BigCharSetModel.objects.exclude(**{lname: "mouldy"})
        assert not_mouldy.count() == 0

    def test_char_len_lookup_empty(self):
        mymodel = BigCharSetModel.objects.create(field=set())

        empty = BigCharSetModel.objects.filter(field__len=0)
        assert empty.count() == 1
        assert empty[0] == mymodel

        one = BigCharSetModel.objects.filter(field__len=1)
        assert one.count() == 0

        one_or_more = BigCharSetModel.objects.filter(field__len__gte=0)
        assert one_or_more.count() == 1

    def test_char_len_lookup(self):
        mymodel = BigCharSetModel.objects.create(field={"red", "expensive"})

        empty = BigCharSetModel.objects.filter(field__len=0)
        assert empty.count() == 0

        one_or_more = BigCharSetModel.objects.filter(field__len__gte=1)
        assert one_or_more.count() == 1
        assert one_or_more[0] == mymodel

        two = BigCharSetModel.objects.filter(field__len=2)
        assert two.count() == 1
        assert two[0] == mymodel

        three = BigCharSetModel.objects.filter(field__len=3)
        assert three.count() == 0

    def test_int_easy(self):
        big_set = {i ** 2 for i in six.moves.range(1000)}
        mymodel = BigIntSetModel.objects.create(field=big_set)
        assert mymodel.field == big_set
        mymodel = BigIntSetModel.objects.get(id=mymodel.id)
        assert mymodel.field == big_set

    def test_int_contains_lookup(self):
        onetwo = BigIntSetModel.objects.create(field={1, 2})

        ones = BigIntSetModel.objects.filter(field__contains=1)
        assert ones.count() == 1
        assert ones[0] == onetwo

        twos = BigIntSetModel.objects.filter(field__contains=2)
        assert twos.count() == 1
        assert twos[0] == onetwo

        threes = BigIntSetModel.objects.filter(field__contains=3)
        assert threes.count() == 0

        with pytest.raises(ValueError):
            list(BigIntSetModel.objects.filter(field__contains={1, 2}))

        ones_and_twos = BigIntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=2)
        )
        assert ones_and_twos.count() == 1
        assert ones_and_twos[0] == onetwo

        ones_and_threes = BigIntSetModel.objects.filter(
            Q(field__contains=1) & Q(field__contains=3)
        )
        assert ones_and_threes.count() == 0

        ones_or_threes = BigIntSetModel.objects.filter(
            Q(field__contains=1) | Q(field__contains=3)
        )
        assert ones_or_threes.count() == 1

        no_three = BigIntSetModel.objects.exclude(field__contains=3)
        assert no_three.count() == 1

        no_one = BigIntSetModel.objects.exclude(field__contains=1)
        assert no_one.count() == 0


class TestValidation(SimpleTestCase):

    def test_max_length(self):
        field = SetTextField(models.CharField(max_length=32), size=3)

        field.clean({'a', 'b', 'c'}, None)

        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean({'a', 'b', 'c', 'd'}, None)
        assert (
            excinfo.value.messages[0] ==
            'Set contains 4 items, it should contain no more than 3.'
        )


class TestCheck(SimpleTestCase):

    def test_model_set(self):
        field = BigIntSetModel._meta.get_field('field')
        assert field.model == BigIntSetModel
        # I think this is a side effect of migrations being run in tests -
        # the base_field.model is the __fake__ model
        assert field.base_field.model.__name__ == 'BigIntSetModel'

    def test_base_field_checks(self):
        class InvalidSetTextModel(TemporaryModel):
            field = SetTextField(models.CharField())

        errors = InvalidSetTextModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E001'
        assert 'Base field for set has errors' in errors[0].msg
        assert 'max_length' in errors[0].msg

    def test_invalid_base_fields(self):
        class InvalidSetTextModel(TemporaryModel):
            field = SetTextField(
                models.ForeignKey('testapp.Author')
            )

        errors = InvalidSetTextModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E002'
        assert 'Base field for set must be' in errors[0].msg


class SetTextFieldSubclass(SetTextField):
    """
    Used below, has a different path for deconstruct()
    """


class TestDeconstruct(TestCase):

    def test_deconstruct(self):
        field = SetTextField(models.IntegerField(), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        assert new.base_field.__class__ == field.base_field.__class__

    def test_deconstruct_with_size(self):
        field = SetTextField(models.IntegerField(), size=3, max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        assert new.size == field.size

    def test_deconstruct_args(self):
        field = SetTextField(models.CharField(max_length=5), max_length=32)
        name, path, args, kwargs = field.deconstruct()
        new = SetTextField(*args, **kwargs)
        assert new.base_field.max_length == field.base_field.max_length

    def test_bad_import_deconstruct(self):
        from django_mysql.models.fields import SetTextField as STField
        field = STField(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.SetTextField'

    def test_bad_import2_deconstruct(self):
        from django_mysql.models.fields.sets import SetTextField as STField
        field = STField(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.SetTextField'

    def test_subclass_deconstruct(self):
        field = SetTextFieldSubclass(models.IntegerField())
        name, path, args, kwargs = field.deconstruct()
        assert path == 'tests.testapp.test_settextfield.SetTextFieldSubclass'


class TestMigrationWriter(TestCase):

    def test_makemigrations(self):
        field = SetTextField(models.CharField(max_length=5))
        statement, imports = MigrationWriter.serialize(field)

        assert (
            statement ==
            "django_mysql.models.SetTextField("
            "models.CharField(max_length=5), "
            "size=None"
            ")"
        )

    def test_makemigrations_with_size(self):
        field = SetTextField(models.CharField(max_length=5), size=5)
        statement, imports = MigrationWriter.serialize(field)

        assert (
            statement ==
            "django_mysql.models.SetTextField("
            "models.CharField(max_length=5), "
            "size=5"
            ")"
        )


class TestSerialization(SimpleTestCase):

    def test_dumping(self):
        big_set = {six.text_type(i ** 2) for i in six.moves.range(1000)}
        instance = BigCharSetModel(field=big_set)
        data = json.loads(serializers.serialize('json', [instance]))[0]
        field = data['fields']['field']
        assert sorted(field.split(',')) == sorted(big_set)

    def test_loading(self):
        test_data = '''
            [{"fields": {"field": "big,leather,comfy"},
             "model": "testapp.BigCharSetModel", "pk": null}]
        '''
        objs = list(serializers.deserialize('json', test_data))
        instance = objs[0].object
        assert instance.field == {"big", "leather", "comfy"}


class TestDescription(SimpleTestCase):

    def test_char(self):
        field = SetTextField(models.CharField(max_length=5), max_length=32)
        assert field.description == "Set of String (up to %(max_length)s)"

    def test_int(self):
        field = SetTextField(models.IntegerField(), max_length=32)
        assert field.description == "Set of Integer"


class TestFormField(SimpleTestCase):

    def test_model_field_formfield(self):
        model_field = SetTextField(models.CharField(max_length=27))
        form_field = model_field.formfield()
        assert isinstance(form_field, SimpleSetField)
        assert isinstance(form_field.base_field, forms.CharField)
        assert form_field.base_field.max_length == 27

    def test_model_field_formfield_size(self):
        model_field = SetTextField(models.IntegerField(), size=400)
        form_field = model_field.formfield()
        assert isinstance(form_field, SimpleSetField)
        assert form_field.max_length == 400



from unittest import SkipTest

import pytest
from django.db import connection, migrations, models, transaction
from django.db.migrations.state import ProjectState
from django.test import TransactionTestCase
from django.test.utils import CaptureQueriesContext

from django_mysql.operations import (
    AlterStorageEngine, InstallPlugin, InstallSOName
)
from django_mysql.test.utils import override_mysql_variables


def plugin_exists(plugin_name):
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT COUNT(*) FROM INFORMATION_SCHEMA.PLUGINS
               WHERE PLUGIN_NAME = %s""",
            (plugin_name,)
        )
        return (cursor.fetchone()[0] > 0)


def table_storage_engine(table_name):
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT ENGINE FROM INFORMATION_SCHEMA.TABLES
               WHERE TABLE_SCHEMA = DATABASE() AND TABLE_NAME = %s""",
            (table_name,)
        )
        return cursor.fetchone()[0]


class PluginOperationTests(TransactionTestCase):

    @classmethod
    def setUpClass(cls):
        super(PluginOperationTests, cls).setUpClass()
        has_metadata_lock_plugin = (
            (connection.is_mariadb and connection.mysql_version >= (10, 0, 7))
        )
        if not has_metadata_lock_plugin:
            raise SkipTest("The metadata_lock_info plugin is required")

    def test_install_plugin(self):
        """
        Test we can load the example plugin that every version of MySQL ships
        with.
        """
        assert not plugin_exists("metadata_lock_info")

        state = ProjectState()
        operation = InstallPlugin("metadata_lock_info",
                                  "metadata_lock_info.so")
        assert (
            operation.describe() ==
            "Installs plugin metadata_lock_info from metadata_lock_info.so"
        )
        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_forwards("testapp", editor,
                                        state, new_state)

        assert plugin_exists("metadata_lock_info")

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_backwards("testapp", editor,
                                         new_state, state)

        assert not plugin_exists("metadata_lock_info")

    def test_install_soname(self):
        """
        Test we can load the 'metadata_lock_info' library.
        """
        assert not plugin_exists("metadata_lock_info")

        state = ProjectState()
        operation = InstallSOName("metadata_lock_info.so")
        assert operation.describe() == "Installs library metadata_lock_info.so"

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_forwards("testapp", editor,
                                        state, new_state)
        assert plugin_exists("metadata_lock_info")

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_backwards("testapp", editor,
                                         new_state, state)
        assert not plugin_exists("metadata_lock_info")


class AlterStorageEngineTests(TransactionTestCase):

    def test_no_from_means_unreversible(self):
        operation = AlterStorageEngine("mymodel", to_engine="InnoDB")
        assert not operation.reversible

        with pytest.raises(NotImplementedError) as excinfo:
            operation.database_backwards(None, None, None, None)

        assert str(excinfo.value) == "You cannot reverse this operation"

    def test_from_means_reversible(self):
        operation = AlterStorageEngine("mymodel", from_engine="MyISAM",
                                       to_engine="InnoDB")
        assert operation.reversible

    def test_describe_without_from(self):
        operation = AlterStorageEngine("Pony", "InnoDB")
        assert (operation.describe() ==
                "Alter storage engine for Pony to InnoDB")

    def test_describe_with_from(self):
        operation = AlterStorageEngine("Pony", from_engine="MyISAM",
                                       to_engine="InnoDB")
        assert (operation.describe() ==
                "Alter storage engine for Pony from MyISAM to InnoDB")

    def test_references_model(self):
        operation = AlterStorageEngine("Pony", "InnoDB")
        assert operation.references_model("PONY")
        assert operation.references_model("Pony")
        assert operation.references_model("pony")
        assert not operation.references_model("poony")

    @override_mysql_variables(default_storage_engine='MyISAM')
    def test_running_with_changes(self):
        project_state = self.set_up_test_model("test_arstd")
        operation = AlterStorageEngine("Pony", from_engine="MyISAM",
                                       to_engine="InnoDB")

        assert table_storage_engine("test_arstd_pony") == "MyISAM"

        # Forwards
        new_state = project_state.clone()
        operation.state_forwards("test_arstd", new_state)
        with connection.schema_editor() as editor:
            operation.database_forwards("test_arstd", editor, project_state,
                                        new_state)
        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Backwards
        with connection.schema_editor() as editor:
            operation.database_backwards("test_arstd", editor, new_state,
                                         project_state)
        assert table_storage_engine("test_arstd_pony") == "MyISAM"

    @override_mysql_variables(default_storage_engine='InnoDB')
    def test_running_without_changes(self):
        project_state = self.set_up_test_model("test_arstd")
        operation = AlterStorageEngine("Pony", from_engine="MyISAM",
                                       to_engine="InnoDB")

        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Forwards - shouldn't actually do an ALTER since it is already InnoDB
        new_state = project_state.clone()
        operation.state_forwards("test_arstd", new_state)
        capturer = CaptureQueriesContext(connection)
        with capturer, connection.schema_editor() as editor:
            operation.database_forwards("test_arstd", editor, project_state,
                                        new_state)
        queries = [q['sql'] for q in capturer.captured_queries]
        assert not any(q.startswith('ALTER TABLE ') for q in queries), (
            "One of the executed queries was an unexpected ALTER TABLE:\n{}"
            .format("\n".join(queries))
        )
        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Backwards - will actually ALTER since it is going 'back' to MyISAM
        with connection.schema_editor() as editor:
            operation.database_backwards("test_arstd", editor, new_state,
                                         project_state)
        assert table_storage_engine("test_arstd_pony") == "MyISAM"

    # Copied from django core migration tests

    def set_up_test_model(
            self, app_label, second_model=False, third_model=False,
            related_model=False, mti_model=False, proxy_model=False,
            unique_together=False, options=False, db_table=None,
            index_together=False):
        """
        Creates a test model state and database table.
        """
        # Delete the tables if they already exist
        table_names = [
            # Start with ManyToMany tables
            '_pony_stables', '_pony_vans',
            # Then standard model tables
            '_pony', '_stable', '_van',
        ]
        tables = [(app_label + table_name) for table_name in table_names]
        with connection.cursor() as cursor:
            table_names = connection.introspection.table_names(cursor)
            connection.disable_constraint_checking()
            sql_delete_table = connection.schema_editor().sql_delete_table
            with transaction.atomic():
                for table in tables:
                    if table in table_names:
                        cursor.execute(sql_delete_table % {
                            "table": connection.ops.quote_name(table),
                        })
            connection.enable_constraint_checking()

        # Make the "current" state
        model_options = {
            "swappable": "TEST_SWAP_MODEL",
            "index_together": [["weight", "pink"]] if index_together else [],
            "unique_together": [["pink", "weight"]] if unique_together else [],
        }
        if options:
            model_options["permissions"] = [("can_groom", "Can groom")]
        if db_table:
            model_options["db_table"] = db_table
        operations = [migrations.CreateModel(
            "Pony",
            [
                ("id", models.AutoField(primary_key=True)),
                ("pink", models.IntegerField(default=3)),
                ("weight", models.FloatField()),
            ],
            options=model_options,
        )]
        if second_model:
            operations.append(migrations.CreateModel(
                "Stable",
                [
                    ("id", models.AutoField(primary_key=True)),
                ]
            ))
        if third_model:
            operations.append(migrations.CreateModel(
                "Van",
                [
                    ("id", models.AutoField(primary_key=True)),
                ]
            ))
        if related_model:
            operations.append(migrations.CreateModel(
                "Rider",
                [
                    ("id", models.AutoField(primary_key=True)),
                    ("pony", models.ForeignKey("Pony")),
                    ("friend", models.ForeignKey("self"))
                ],
            ))
        if mti_model:
            operations.append(migrations.CreateModel(
                "ShetlandPony",
                fields=[
                    ('pony_ptr', models.OneToOneField(
                        auto_created=True,
                        primary_key=True,
                        to_field='id',
                        serialize=False,
                        to='Pony',
                    )),
                    ("cuteness", models.IntegerField(default=1)),
                ],
                bases=['%s.Pony' % app_label],
            ))
        if proxy_model:
            operations.append(migrations.CreateModel(
                "ProxyPony",
                fields=[],
                options={"proxy": True},
                bases=['%s.Pony' % app_label],
            ))

        return self.apply_operations(app_label, ProjectState(), operations)

    def apply_operations(self, app_label, project_state, operations):
        migration = migrations.Migration('name', app_label)
        migration.operations = operations
        with connection.schema_editor() as editor:
            return migration.apply(project_state, editor)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from optparse import make_option

import django
from django.core.management import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.utils import ConnectionDoesNotExist

from django_mysql.utils import settings_to_cmd_args


class Command(BaseCommand):
    args = "<optional connection alias>"

    help = ("Outputs shell parameters representing database connection "
            "suitable for inclusion in various tools' commandlines. The "
            "connection alias should be a name from DATABASES - defaults to "
            "'{default}'.").format(default=DEFAULT_DB_ALIAS)

    if django.VERSION[:2] < (1, 8):

        option_list = BaseCommand.option_list + (
            make_option(
                '--mysql',
                action='store_true',
                dest='mysql',
                default=False,
                help='Outputs flags for tools that take parameters in the '
                     'same format as the mysql client, e.g. mysql '
                     '$(./manage.py dbparams --mysql)'
            ),
            make_option(
                '--dsn',
                action='store_true',
                dest='dsn',
                default=False,
                help='Output a DSN for e.g. percona tools, e.g. '
                     'pt-online-schema-change $(./manage.py dbparams --dsn)'
            ),
        )

    else:

        def add_arguments(self, parser):
            if django.VERSION[:2] >= (1, 10):
                parser.add_argument(
                    'alias', metavar='alias', nargs='?',
                    default=DEFAULT_DB_ALIAS,
                    help='Specify the database connection alias to output '
                         'parameters for.'
                )

            parser.add_argument(
                '--mysql',
                action='store_true',
                dest='mysql',
                default=False,
                help='Outputs flags for tools that take parameters in the '
                     'same format as the mysql client, e.g. mysql '
                     '$(./manage.py dbparams --mysql)'
            )
            parser.add_argument(
                '--dsn',
                action='store_true',
                dest='dsn',
                default=False,
                help='Output a DSN for e.g. percona tools, e.g. '
                     'pt-online-schema-change $(./manage.py dbparams --dsn)'
            ),

    def handle(self, *args, **options):
        if django.VERSION[:2] >= (1, 10):
            alias = options['alias']
        else:
            if len(args) > 1:
                raise CommandError('Cannot output the parameters for more '
                                   'than one connection.')
            elif len(args) == 0:
                alias = DEFAULT_DB_ALIAS
            else:
                alias = args[0]

        try:
            settings_dict = connections[alias].settings_dict
        except ConnectionDoesNotExist:
            raise CommandError("Connection '{}' does not exist".format(alias))

        connection = connections[alias]
        if connection.vendor != 'mysql':
            raise CommandError("{} is not a MySQL database connection"
                               .format(alias))

        show_mysql = options['mysql']
        show_dsn = options['dsn']
        if show_mysql and show_dsn:
            raise CommandError("Pass only one of --mysql and --dsn")
        elif not show_mysql and not show_dsn:
            show_mysql = True

        if show_mysql:
            self.output_for_mysql(settings_dict)
        elif show_dsn:
            self.output_for_dsn(settings_dict)

    def output_for_mysql(self, settings_dict):
        args = settings_to_cmd_args(settings_dict)
        args = args[1:]  # Delete the 'mysql' at the start
        self.stdout.write(" ".join(args), ending="")

    def output_for_dsn(self, settings_dict):
        cert = settings_dict['OPTIONS'].get('ssl', {}).get('ca')
        if cert:
            self.stderr.write(
                "Warning: SSL params can't be passed in the DSN syntax; you "
                "must pass them in your my.cnf. See: "
                "http://www.percona.com/blog/2014/10/16/percona-toolkit-for-"
                "mysql-with-mysql-ssl-connections/"
            )

        db = settings_dict['OPTIONS'].get('db', settings_dict['NAME'])
        user = settings_dict['OPTIONS'].get('user', settings_dict['USER'])
        passwd = settings_dict['OPTIONS'].get('passwd',
                                              settings_dict['PASSWORD'])
        host = settings_dict['OPTIONS'].get('host', settings_dict['HOST'])
        port = settings_dict['OPTIONS'].get('port', settings_dict['PORT'])
        defaults_file = settings_dict['OPTIONS'].get('read_default_file')

        args = []
        if defaults_file:
            args.append('F={}'.format(defaults_file))
        if user:
            args.append('u={}'.format(user))
        if passwd:
            args.append('p={}'.format(passwd))
        if host:
            if '/' in host:
                args.append('S={}'.format(host))
            else:
                args.append('h={}'.format(host))
        if port:
            args.append('P={}'.format(port))
        if db:
            args.append('D={}'.format(db))

        dsn = ",".join(args)
        self.stdout.write(dsn, ending="")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

from django_mysql.models import SizedBinaryField


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sizedbinaryaltermodel',
            name='field',
            field=SizedBinaryField(size_class=2)
        ),
    ]



# -*- coding:utf-8 -*-
from unittest import skipIf

import django
from django.core.management import call_command
from django.test import TestCase, TransactionTestCase

from django_mysql.checks import check_variables
from django_mysql.test.utils import override_mysql_variables

requiresNoDBConnection = skipIf(
    django.VERSION[:2] < (1, 8),
    "Requires nodb_connection from Django 1.8+"
)


class CallCheckTest(TestCase):

    def test_check(self):
        call_command('check')


@requiresNoDBConnection
class VariablesTests(TransactionTestCase):

    def test_passes(self):
        assert check_variables([]) == []

    @override_mysql_variables(sql_mode="")
    def test_fails_if_no_strict(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.W001'
        assert "MySQL Strict Mode" in errors[0].msg

    @override_mysql_variables(innodb_strict_mode=0)
    def test_fails_if_no_innodb_strict(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.W002'
        assert "InnoDB Strict Mode" in errors[0].msg

    @override_mysql_variables(character_set_connection='utf8')
    def test_fails_if_not_utf8mb4(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.W003'
        assert "utf8mb4" in errors[0].msg



# -*- coding:utf-8 -*-
import pytest
from django import forms
from django.core import exceptions
from django.test import SimpleTestCase

from django_mysql.forms import JSONField, SimpleListField, SimpleSetField


class TestSimpleListField(SimpleTestCase):

    def test_valid(self):
        field = SimpleListField(forms.CharField())
        value = field.clean('a,b,c')
        assert value == ['a', 'b', 'c']

    def test_to_python_no_leading_commas(self):
        field = SimpleListField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean(',1')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_trailing_commas(self):
        field = SimpleListField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_double_commas(self):
        field = SimpleListField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,,2')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_base_field_does_not_validate(self):
        field = SimpleListField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,9')
        assert (
            excinfo.value.messages[0] ==
            'Item 1 in the list did not validate: Enter a whole number.'
        )

    def test_validate_fail(self):
        field = SimpleListField(
            forms.ChoiceField(choices=(('a', 'The letter A'),
                                       ('b', 'The letter B')))
        )
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,c')
        assert (
            excinfo.value.messages[0] ==
            'Item 2 in the list did not validate: '
            'Select a valid choice. c is not one of the available choices.'
        )

    def test_validators_fail(self):
        field = SimpleListField(forms.RegexField('[a-e]{2}'))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,bc,de')
        assert (
            excinfo.value.messages[0] ==
            'Item 1 in the list did not validate: Enter a valid value.'
        )

    def test_validators_fail_base_max_length(self):
        field = SimpleListField(forms.CharField(max_length=5))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('longer,yes')
        assert (
            excinfo.value.messages[0] ==
            'Item 1 in the list did not validate: '
            'Ensure this value has at most 5 characters (it has 6).'
        )

    def test_validators_fail_base_min_max_length(self):
        # there's just no satisfying some people...
        field = SimpleListField(forms.CharField(min_length=10, max_length=8))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('undefined')
        assert (
            excinfo.value.messages[0] ==
            'Item 1 in the list did not validate: '
            'Ensure this value has at least 10 characters (it has 9).'
        )
        assert (
            excinfo.value.messages[1] ==
            'Item 1 in the list did not validate: '
            'Ensure this value has at most 8 characters (it has 9).'
        )

    def test_prepare_value(self):
        field = SimpleListField(forms.CharField())
        value = field.prepare_value(['a', 'b', 'c'])
        assert value.split(',') == ['a', 'b', 'c']

        assert field.prepare_value('1,a') == '1,a'

    def test_max_length(self):
        field = SimpleListField(forms.CharField(), max_length=2)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,c')
        assert (
            excinfo.value.messages[0] ==
            'List contains 3 items, it should contain no more than 2.'
        )

    def test_min_length(self):
        field = SimpleListField(forms.CharField(), min_length=4)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,c')
        assert (
            excinfo.value.messages[0] ==
            'List contains 3 items, it should contain no fewer than 4.'
        )

    def test_required(self):
        field = SimpleListField(forms.CharField(), required=True)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('')
        assert excinfo.value.messages[0] == 'This field is required.'


class TestSimpleSetField(SimpleTestCase):

    def test_valid(self):
        field = SimpleSetField(forms.CharField())
        value = field.clean('a,b,c')
        assert value == {'a', 'b', 'c'}

    def test_to_python_no_leading_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean(',1')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_trailing_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_no_double_commas(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,,2')
        assert (
            excinfo.value.messages[0] ==
            'No leading, trailing, or double commas.'
        )

    def test_to_python_base_field_does_not_validate(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,9')
        assert (
            excinfo.value.messages[0] ==
            'Item 1 in the set did not validate: Enter a whole number.'
        )

    def test_to_python_duplicates_not_allowed(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,1')
        assert (
            excinfo.value.messages[0] ==
            "Duplicates are not supported. '1' appears twice or more."
        )

    def test_to_python_two_duplicates_not_allowed(self):
        field = SimpleSetField(forms.IntegerField())
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('1,2,1,2')
        assert (
            excinfo.value.messages[0] ==
            "Duplicates are not supported. '1' appears twice or more."
        )
        assert (
            excinfo.value.messages[1] ==
            "Duplicates are not supported. '2' appears twice or more."
        )

    def test_validate_fail(self):
        field = SimpleSetField(
            forms.ChoiceField(choices=(('a', 'The letter A'),
                                       ('b', 'The letter B')))
        )
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,c')
        assert (
            excinfo.value.messages[0] ==
            'Item "c" in the set did not validate: '
            'Select a valid choice. c is not one of the available choices.'
        )

    def test_validators_fail(self):
        field = SimpleSetField(forms.RegexField('[a-e]{2}'))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,bc,de')
        assert (
            excinfo.value.messages[0] ==
            'Item "a" in the set did not validate: Enter a valid value.'
        )

    def test_validators_fail_base_max_length(self):
        field = SimpleSetField(forms.CharField(max_length=5))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('longer,yes')
        assert (
            excinfo.value.messages[0] ==
            'Item "longer" in the set did not validate: '
            'Ensure this value has at most 5 characters (it has 6).'
        )

    def test_validators_fail_base_min_max_length(self):
        # there's just no satisfying some people...
        field = SimpleSetField(forms.CharField(min_length=10, max_length=8))
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('undefined')
        assert (
            excinfo.value.messages[0] ==
            'Item "undefined" in the set did not validate: '
            'Ensure this value has at least 10 characters (it has 9).'
        )
        assert (
            excinfo.value.messages[1] ==
            'Item "undefined" in the set did not validate: '
            'Ensure this value has at most 8 characters (it has 9).'
        )

    def test_prepare_value(self):
        field = SimpleSetField(forms.CharField())
        value = field.prepare_value({'a', 'b', 'c'})
        assert (
            list(sorted(value.split(','))) ==
            ['a', 'b', 'c']
        )

        assert field.prepare_value('1,a') == '1,a'

    def test_max_length(self):
        field = SimpleSetField(forms.CharField(), max_length=2)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,c')
        assert (
            excinfo.value.messages[0] ==
            'Set contains 3 items, it should contain no more than 2.'
        )

    def test_min_length(self):
        field = SimpleSetField(forms.CharField(), min_length=4)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('a,b,c')
        assert (
            excinfo.value.messages[0] ==
            'Set contains 3 items, it should contain no fewer than 4.'
        )

    def test_required(self):
        field = SimpleSetField(forms.CharField(), required=True)
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('')
        assert excinfo.value.messages[0] == 'This field is required.'


class TestJSONField(SimpleTestCase):

    def test_valid(self):
        field = JSONField()
        value = field.clean('{"a": "b"}')
        assert value == {'a': 'b'}

    def test_valid_empty(self):
        field = JSONField(required=False)
        value = field.clean('')
        assert value is None

    def test_invalid(self):
        field = JSONField()
        with pytest.raises(exceptions.ValidationError) as excinfo:
            field.clean('{some badly formed: json}')
        assert (
            excinfo.value.messages[0] ==
            "'{some badly formed: json}' value must be valid JSON."
        )

    def test_prepare_value(self):
        field = JSONField()
        assert field.prepare_value({'a': 'b'}) == '{"a": "b"}'
        assert field.prepare_value(None) == 'null'



# -*- coding:utf-8 -*-
from __future__ import unicode_literals

import json
import mock
from datetime import date, datetime, time
from unittest import SkipTest, skipIf

import django
import pytest
from django.core import serializers
from django.db import connection, connections
from django.db.migrations.writer import MigrationWriter
from django.db.models import CharField, Transform
from django.test import TestCase
from django.utils import six

from django_mysql.models import DynamicField
from testapp.models import DynamicModel, TemporaryModel
from testapp.utils import requiresPython2


class DynColTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        if not (
            connection.is_mariadb and
            connection.mysql_version >= (10, 0, 1)
        ):
            raise SkipTest("Dynamic Columns require MariaDB 10.0.1+")
        super(DynColTestCase, cls).setUpClass()


class TestSaveLoad(DynColTestCase):

    def test_save_and_mutations(self):
        s = DynamicModel.objects.create()
        assert s.attrs == {}

        s = DynamicModel.objects.get()
        assert s.attrs == {}

        s.attrs['key'] = 'value!'
        s.attrs['2key'] = 23
        s.save()
        s = DynamicModel.objects.get()
        assert s.attrs == {'key': 'value!', '2key': 23}

        del s.attrs['key']
        s.save()
        s = DynamicModel.objects.get()
        assert s.attrs == {'2key': 23}

        del s.attrs['2key']
        s.save()
        s = DynamicModel.objects.get()
        assert s.attrs == {}

    def test_create(self):
        DynamicModel.objects.create(attrs={
            'a': 'value'
        })
        s = DynamicModel.objects.get()
        assert s.attrs == {'a': 'value'}

    def test_create_succeeds_specced_field(self):
        DynamicModel.objects.create(attrs={'inty': 1})
        s = DynamicModel.objects.get()
        assert s.attrs == {'inty': 1}

    def test_create_fails_bad_value(self):
        with pytest.raises(TypeError):
            DynamicModel.objects.create(attrs={'inty': 1.0})

    def test_bulk_create(self):
        DynamicModel.objects.bulk_create([
            DynamicModel(attrs={'a': 'value'}),
            DynamicModel(attrs={'b': 'value2'}),
        ])
        dm1, dm2 = DynamicModel.objects.all().order_by('id')
        assert dm1.attrs == {'a': 'value'}
        assert dm2.attrs == {'b': 'value2'}


class SpecTests(DynColTestCase):

    def test_spec_dict_type(self):
        DynamicField.validate_spec(
            {'a': dict},
            {'a': {'this': 'that'}}
        )  # no errors

    def test_illegal_int(self):
        m = DynamicModel(attrs={'inty': 1.0})
        with pytest.raises(TypeError) as excinfo:
            m.save()
        assert "Key 'inty' should be of type int" in str(excinfo.value)

    def test_illegal_nested(self):
        m = DynamicModel(attrs={'nesty': {'level2': 1}})
        with pytest.raises(TypeError) as excinfo:
            m.save()
        assert "Key 'nesty.level2' should be of type " in str(excinfo.value)

    def test_illegal_nested_type(self):
        m = DynamicModel(attrs={'nesty': []})
        with pytest.raises(TypeError) as excinfo:
            m.save()
        assert "Key 'nesty' should be of type dict" in str(excinfo.value)

    @requiresPython2
    def test_long_equivalent_to_int(self):
        from __builtin__ import long  # make source lintable on Python 3
        DynamicField.validate_spec({'a': int}, {'a': long(9001)})  # no errors

    @requiresPython2
    def test_int_equivalent_to_long(self):
        from __builtin__ import long  # make source lintable on Python 3
        DynamicField.validate_spec({'a': long}, {'a': int(9001)})  # no errors


class DumbTransform(Transform):
    """
    Used to test existing transform behaviour - by default in Django there are
    no transforms on BinaryField.
    Really dumb, returns the string 'dumb' always
    """
    lookup_name = 'dumb'
    output_field = CharField()

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return "%s", ['dumb']

DynamicField.register_lookup(DumbTransform)


class QueryTests(DynColTestCase):

    def setUp(self):
        super(QueryTests, self).setUp()
        self.objs = [
            DynamicModel(attrs={'a': 'b'}),
            DynamicModel(attrs={'a': 'b', 'c': 'd'}),
            DynamicModel(attrs={'c': 'd'}),
            DynamicModel(attrs={}),
            DynamicModel(attrs={
                'datetimey': datetime(2001, 1, 4, 14, 15, 16),
                'datey': date(2001, 1, 4),
                'floaty': 128.5,
                'inty': 9001,
                'stry': "strvalue",
                'str_underscorey': "strvalue2",
                'timey': time(14, 15, 16),
                'nesty': {
                    'level2': 'chirp'
                }
            }),
        ]
        DynamicModel.objects.bulk_create(self.objs)
        self.objs = list(DynamicModel.objects.all().order_by('id'))

    def test_equal(self):
        assert (
            list(DynamicModel.objects.filter(attrs={'a': 'b'})) ==
            self.objs[:1]
        )

    def test_exact(self):
        assert (
            list(DynamicModel.objects.filter(attrs__exact={'a': 'b'})) ==
            self.objs[:1]
        )

    def test_preexisting_transforms_work_fine(self):
        assert (
            list(DynamicModel.objects.filter(attrs__dumb='notdumb')) ==
            []
        )

    def test_has_key(self):
        assert (
            list(DynamicModel.objects.filter(attrs__has_key='c')) ==
            self.objs[1:3]
        )

    def test_key_transform_datey(self):
        assert (
            list(DynamicModel.objects.filter(attrs__datey=date(2001, 1, 4))) ==
            [self.objs[4]]
        )

    def test_key_transform_datey_DATE(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__datey_DATE=date(2001, 1, 4)
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_datetimey(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__datetimey=datetime(2001, 1, 4, 14, 15, 16)
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_datetimey__year(self):
        assert (
            list(DynamicModel.objects.filter(attrs__datetimey__year=2001)) ==
            [self.objs[4]]
        )

    def test_key_transform_datetimey_DATETIME(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__datetimey_DATETIME=datetime(2001, 1, 4, 14, 15, 16)
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_floaty(self):
        assert (
            list(DynamicModel.objects.filter(attrs__floaty__gte=128.0)) ==
            [self.objs[4]]
        )

    def test_key_transform_floaty_DOUBLE(self):
        assert (
            list(DynamicModel.objects.filter(attrs__floaty_DOUBLE=128.5)) ==
            [self.objs[4]]
        )

    def test_key_transform_inty(self):
        assert (
            list(DynamicModel.objects.filter(attrs__inty=9001)) ==
            [self.objs[4]]
        )

    def test_key_transform_inty_INTEGER(self):
        assert (
            list(DynamicModel.objects.filter(attrs__inty_INTEGER=9001)) ==
            [self.objs[4]]
        )

    def test_key_transform_inty_no_results(self):
        assert (
            list(DynamicModel.objects.filter(attrs__inty=12991)) ==
            []
        )

    def test_key_transform_inty_in_subquery(self):
        assert (
            list(DynamicModel.objects.filter(
                id__in=DynamicModel.objects.filter(attrs__inty=9001),
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_miss_CHAR_isnull(self):
        assert (
            list(DynamicModel.objects.filter(attrs__miss_CHAR__isnull=True)) ==
            self.objs
        )

    def test_key_transform_stry(self):
        assert (
            list(DynamicModel.objects.filter(attrs__stry="strvalue")) ==
            [self.objs[4]]
        )

    def test_key_transform_stry_CHAR(self):
        assert (
            list(DynamicModel.objects.filter(attrs__stry_CHAR="strvalue")) ==
            [self.objs[4]]
        )

    def test_key_transform_str_underscorey_CHAR(self):
        # Check that underscores in key names are parsed fine
        assert (
            list(DynamicModel.objects.filter(
                attrs__str_underscorey_CHAR="strvalue2"
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_timey(self):
        assert (
            list(DynamicModel.objects.filter(attrs__timey=time(14, 15, 16))) ==
            [self.objs[4]]
        )

    def test_key_transform_timey_TIME(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__timey_TIME=time(14, 15, 16)
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_nesty__level2(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__nesty__level2='chirp'
            )) ==
            [self.objs[4]]
        )

    def test_key_transform_nesty__level2__startswith(self):
        assert (
            list(DynamicModel.objects.filter(
                attrs__nesty__level2__startswith='chi'
            )) ==
            [self.objs[4]]
        )

requiresDBChecks = skipIf(
    django.VERSION[:2] < (1, 8),
    "Requires ability to talk to DB in checks from Django 1.8+"
)


class TestCheck(DynColTestCase):

    wrapper_path = 'django.db.backends.mysql.base.DatabaseWrapper'

    @requiresDBChecks
    @mock.patch(wrapper_path + '.is_mariadb', new=False)
    def test_db_not_mariadb(self):
        # Uncache cached_property
        for db in connections:
            if 'is_mariadb' in connections[db].__dict__:
                del connections[db].__dict__['is_mariadb']

        class ValidDynamicModel(TemporaryModel):
            field = DynamicField()

        errors = ValidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E013'
        assert "MariaDB 10.0.1+ is required" in errors[0].msg

    @requiresDBChecks
    @mock.patch(wrapper_path + '.mysql_version', new=(5, 5, 3))
    def test_mariadb_old_version(self):
        # Uncache cached_property
        for db in connections:
            if 'mysql_version' in connections[db].__dict__:
                del connections[db].__dict__['mysql_version']

        class ValidDynamicModel(TemporaryModel):
            field = DynamicField()

        errors = ValidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E013'
        assert "MariaDB 10.0.1+ is required" in errors[0].msg

    @requiresDBChecks
    @mock.patch(DynamicField.__module__ + '.mariadb_dyncol', new=None)
    def test_mariadb_dyncol_missing(self):
        errors = DynamicModel.check()
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E012'
        assert "'mariadb_dyncol' is required" in errors[0].msg

    @requiresDBChecks
    def test_character_set_not_utf8_compatible(self):
        with connection.cursor() as cursor:
            cursor.execute("SELECT @@character_set_client")
            orig_charset = cursor.fetchone()[0]
            cursor.execute("SET NAMES 'latin1'")
            try:
                errors = DynamicModel.check()
            finally:
                cursor.execute("SET NAMES '{}'".format(orig_charset))

        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E014'
        assert "The MySQL charset must be 'utf8'" in errors[0].msg

    def test_spec_not_dict(self):
        class InvalidDynamicModel(TemporaryModel):
            field = DynamicField(spec=['woops', 'a', 'list'])

        errors = InvalidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E009'
        assert "'spec' must be a dict" in errors[0].msg
        assert "The value passed is of type list" in errors[0].hint

    def test_spec_key_not_valid(self):
        class InvalidDynamicModel(TemporaryModel):
            field = DynamicField(spec={
                2.0: six.text_type
            })

        errors = InvalidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E010'
        assert "The key '2.0' in 'spec' is not a string" in errors[0].msg
        assert "'spec' keys must be of type " in errors[0].hint
        assert "'2.0' is of type float" in errors[0].hint

    def test_spec_value_not_valid(self):
        class InvalidDynamicModel(TemporaryModel):
            field = DynamicField(spec={'bad': list})

        errors = InvalidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E011'
        assert (
            "The value for 'bad' in 'spec' is not an allowed type" in
            errors[0].msg
        )
        assert (
            "'spec' values must be one of the following types: "
            "date, datetime" in
            errors[0].hint
        )

    def test_spec_nested_value_not_valid(self):
        class InvalidDynamicModel(TemporaryModel):
            field = DynamicField(spec={
                'l1': {
                    'bad': tuple
                }
            })

        errors = InvalidDynamicModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == 'django_mysql.E011'
        assert (
            "The value for 'bad' in 'spec.l1' is not an allowed type" in
            errors[0].msg
        )
        assert (
            "'spec' values must be one of the following types: "
            "date, datetime" in
            errors[0].hint
        )


class SubDynamicField(DynamicField):
    """
    Used below, has a different path for deconstruct()
    """


class TestDeconstruct(TestCase):

    def test_deconstruct(self):
        field = DynamicField()
        name, path, args, kwargs = field.deconstruct()
        DynamicField(*args, **kwargs)

    def test_deconstruct_spec(self):
        field = DynamicField(spec={'this': int, 'that': float})
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.DynamicField'
        DynamicField(*args, **kwargs)

    def test_bad_import_deconstruct(self):
        from django_mysql.models.fields import DynamicField as DField
        field = DField()
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.DynamicField'

    def test_bad_import2_deconstruct(self):
        from django_mysql.models.fields.dynamic import DynamicField as DField
        field = DField()
        name, path, args, kwargs = field.deconstruct()
        assert path == 'django_mysql.models.DynamicField'

    def test_subclass_deconstruct(self):
        field = SubDynamicField()
        name, path, args, kwargs = field.deconstruct()
        assert path == 'tests.testapp.test_dynamicfield.SubDynamicField'


class TestMigrations(DynColTestCase):

    def test_makemigrations(self):
        field = DynamicField(spec={'a': 'beta'})
        statement, imports = MigrationWriter.serialize(field)
        # 'spec' should not appear since that would trigger needless ALTERs
        assert statement == "django_mysql.models.DynamicField()"


class TestSerialization(DynColTestCase):
    test_data = '''
        [{"fields": {"attrs": "{\\"a\\": \\"b\\"}"},
          "model": "testapp.dynamicmodel", "pk": null}]
    '''

    def test_dumping(self):
        instance = DynamicModel(attrs={'a': 'b'})
        data = serializers.serialize('json', [instance])
        assert json.loads(data) == json.loads(self.test_data)

    def test_loading(self):
        deserialized = list(serializers.deserialize('json', self.test_data))
        instance = deserialized[0].object
        assert instance.attrs == {'a': 'b'}


class TestFormfield(DynColTestCase):
    def test_formfield(self):
        model_field = DynamicField()
        form_field = model_field.formfield()
        self.assertIsNone(form_field)



# -*- coding:utf-8 -*-
from django.db import connection
from django.test import TestCase
from django.test.utils import override_settings

from django_mysql.monkey_patches import patch_CursorWrapper_execute
from django_mysql.rewrite_query import rewrite_query
from testapp.utils import CaptureLastQuery


class RewriteQueryTests(TestCase):

    def test_it_doesnt_touch_normal_queries(self):
        self.check_identity("SELECT 1")
        self.check_identity("SELECT col_a, col_b FROM sometable WHERE 1")

    def test_bad_rewrites_ignored(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAY_JOIN*/1)"
            ) ==
            "SELECT col_a FROM sometable WHERE (1)"
        )
        assert (
            rewrite_query(
                "SELECT col_a FROM sometable "
                "WHERE (/*QueryRewrite':*/1)"
            ) ==
            "SELECT col_a FROM sometable WHERE (1)"
        )
        assert (
            rewrite_query(
                "UPDATE col_a SET pants='onfire' "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "UPDATE col_a SET pants='onfire' WHERE (1)"
        )

    def test_non_select_update_deletes_ignored(self):
        assert (
            rewrite_query("SHOW TABLES /*QueryRewrite':STRAIGHT_JOIN*/") ==
            "SHOW TABLES "
        )

    def check_identity(self, query):
        assert rewrite_query(query) == query

    def test_straight_join(self):
        assert (
            rewrite_query(
                "SELECT col_a, col_b FROM sometable "
                "WHERE nothing() AND (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT STRAIGHT_JOIN col_a, col_b FROM sometable "
            "WHERE nothing() AND (1)"
        )

    def test_straight_join_preceeding_whitespace(self):
        assert (
            rewrite_query(
                "  SELECT col_a, col_b FROM sometable "
                "WHERE nothing() AND (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT STRAIGHT_JOIN col_a, col_b FROM sometable "
            "WHERE nothing() AND (1)"
        )

    def test_straight_join_with_comment(self):
        assert (
            rewrite_query(
                "SELECT /* HI MUM */ col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT /* HI MUM */ STRAIGHT_JOIN col_a FROM sometable "
            "WHERE (1)"
        )

    def test_straight_join_with_comments(self):
        assert (
            rewrite_query(
                "SELECT /* GOODBYE */ col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT /* GOODBYE */ STRAIGHT_JOIN col_a FROM "
            "sometable WHERE (1)"
        )

    def test_straight_join_with_repeat_comments(self):
        assert (
            rewrite_query(
                "SELECT /* A */ /* B */ /* C */ col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT /* A */ /* B */ /* C */ STRAIGHT_JOIN col_a FROM "
            "sometable WHERE (1)"
        )

    def test_straight_join_with_spaceless_comment(self):
        assert (
            rewrite_query(
                "SELECT/* this*/col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT /* this*/ STRAIGHT_JOIN col_a FROM sometable WHERE (1)"
        )

    def test_straight_join_idempotent(self):
        assert (
            rewrite_query(
                "SELECT col_a, col_b FROM sometable "
                "WHERE nothing() AND (/*QueryRewrite':STRAIGHT_JOIN*/1) "
                "AND (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT STRAIGHT_JOIN col_a, col_b FROM sometable "
            "WHERE nothing() AND (1) AND (1)"
        )

    def test_straight_join_doesnt_affect_distinct(self):
        assert (
            rewrite_query(
                "SELECT DISTINCT col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT DISTINCT STRAIGHT_JOIN col_a FROM sometable WHERE (1)"
        )

    def test_straight_join_doesnt_affect_all_and_highpriority(self):
        assert (
            rewrite_query(
                "SELECT ALL HIGH_PRIORITY col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT ALL HIGH_PRIORITY STRAIGHT_JOIN col_a FROM sometable "
            "WHERE (1)"
        )

    def test_2_straight_joins_dont_affect_all_and_highpriority(self):
        assert (
            rewrite_query(
                "SELECT ALL HIGH_PRIORITY col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1) AND "
                "(/*QueryRewrite':STRAIGHT_JOIN*/1)"
            ) ==
            "SELECT ALL HIGH_PRIORITY STRAIGHT_JOIN col_a FROM sometable "
            "WHERE (1) AND (1)"
        )

    def test_multiple_hints(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM sometable "
                "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1) AND "
                "(/*QueryRewrite':SQL_NO_CACHE*/1)"
            ) ==
            "SELECT STRAIGHT_JOIN SQL_NO_CACHE col_a FROM sometable "
            "WHERE (1) AND (1)"
        )

    def test_mutually_exclusive_latest_wins(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM sometable "
                "WHERE (/*QueryRewrite':SQL_CACHE*/1) AND "
                "(/*QueryRewrite':SQL_NO_CACHE*/1)"
            ) ==
            "SELECT SQL_NO_CACHE col_a FROM sometable WHERE (1) AND (1)"
        )

    def test_labelling(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM sometable "
                "WHERE (/*QueryRewrite':label=himum*/1)"
            ) ==
            "SELECT /*himum*/ col_a FROM sometable WHERE (1)"
        )

    def test_labelling_mysql_57_hint(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM t1 "
                "WHERE (/*QueryRewrite':label=+ NO_RANGE_OPTIMIZATION(t1 "
                "PRIMARY) */1)"
            ) ==
            "SELECT /*+ NO_RANGE_OPTIMIZATION(t1 PRIMARY) */ col_a FROM t1 "
            "WHERE (1)"
        )

    def test_not_case_sensitive(self):
        assert (
            rewrite_query(
                "select col_a from sometable "
                "where (/*QueryRewrite':label=himum*/1)"
            ) ==
            "select /*himum*/ col_a from sometable "
            "where (1)"
        )

    def test_bad_query_not_rewritten(self):
        assert (
            rewrite_query(
                "SELECTSTRAIGHT_JOIN col_a FROM sometable"
                "WHERE (/*QueryRewrite':label=hi*/1)"
            ) ==
            "SELECTSTRAIGHT_JOIN col_a FROM sometable"
            "WHERE (1)"
        )

    def test_index_hint_use(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` WHERE "
                "(/*QueryRewrite':index=`sometable` USE `col_a_idx`*/1)"
            ) ==
            "SELECT col_a FROM `sometable` USE INDEX (`col_a_idx`) WHERE (1)"
        )

    def test_index_ignore(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` WHERE "
                "(/*QueryRewrite':index=`sometable` IGNORE `col_a_idx`*/1)"
            ) ==
            "SELECT col_a FROM `sometable` IGNORE INDEX (`col_a_idx`) "
            "WHERE (1)"
        )

    def test_index_force(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` WHERE "
                "(/*QueryRewrite':index=`sometable` FORCE `col_a_idx`*/1)"
            ) ==
            "SELECT col_a FROM `sometable` FORCE INDEX (`col_a_idx`) "
            "WHERE (1)"
        )

    def test_index_nonsense_does_nothing(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` WHERE "
                "(/*QueryRewrite':index=`sometable` MAHOGANY `col_a_idx`*/1)"
            ) ==
            "SELECT col_a FROM `sometable` WHERE (1)"
        )

    def test_index_hint_use_secondary(self):
        assert (
            rewrite_query(
                "SELECT col_a, col_b FROM `sometable` INNER JOIN `othertable` "
                " WHERE (/*QueryRewrite':index=`othertable` USE `myindex`*/1)"
            ) ==
            "SELECT col_a, col_b FROM `sometable` INNER JOIN `othertable` "
            "USE INDEX (`myindex`) WHERE (1)"
        )

    def test_index_hint_multiple_indexes(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `tabl` WHERE "
                "(/*QueryRewrite':index=`tabl` IGNORE `idx1`,`idx2`*/1)"
            ) ==
            "SELECT col_a FROM `tabl` IGNORE INDEX (`idx1`,`idx2`) WHERE (1)"
        )

    def test_index_hint_multiple_hints(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` "
                "WHERE (/*QueryRewrite':index=`sometable` IGNORE `idx1`*/1) "
                "AND (/*QueryRewrite':index=`sometable` IGNORE `idx2`*/1)"
            ) ==
            "SELECT col_a FROM `sometable` IGNORE INDEX (`idx2`) "
            "IGNORE INDEX (`idx1`) WHERE (1) AND (1)"
        )

    def test_index_hint_for_join(self):
        assert (
            rewrite_query(
                "SELECT `sometable`.col_a, `sometable2`.col_b "
                "FROM `sometable` NATURAL JOIN `sometable2` "
                "WHERE (/*QueryRewrite':index=`sometable` IGNORE FOR JOIN "
                "`idx`*/1)"
            ) ==
            "SELECT `sometable`.col_a, `sometable2`.col_b FROM `sometable` "
            "IGNORE INDEX FOR JOIN (`idx`) "
            "NATURAL JOIN `sometable2` WHERE (1)"
        )

    def test_index_hint_for_group_by(self):
        assert (
            rewrite_query(
                "SELECT col_a, SUM(col_b) FROM `sometable` "
                "WHERE (/*QueryRewrite':index=`sometable` FORCE FOR GROUP BY "
                "`idx`*/1) GROUP BY col_a"
            ) ==
            "SELECT col_a, SUM(col_b) FROM `sometable` FORCE INDEX FOR GROUP "
            "BY (`idx`) WHERE (1) GROUP BY col_a"
        )

    def test_index_hint_for_order_by(self):
        assert (
            rewrite_query(
                "SELECT col_a FROM `sometable` "
                "WHERE (/*QueryRewrite':index=`sometable` USE FOR ORDER BY "
                "`idx` */1) ORDER BY col_a"
            ) ==
            "SELECT col_a FROM `sometable` USE INDEX FOR ORDER BY (`idx`) "
            "WHERE (1) ORDER BY col_a"
        )

    def test_it_is_monkey_patched(self):
        with CaptureLastQuery() as cap, connection.cursor() as cursor:
            cursor.execute("SELECT 1 FROM DUAL "
                           "WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)")
        assert cap.query == "SELECT STRAIGHT_JOIN 1 FROM DUAL WHERE (1)"

    @override_settings(DJANGO_MYSQL_REWRITE_QUERIES=False)
    def test_monkey_patch_can_be_disabled(self):
        query = "SELECT 1 FROM DUAL WHERE (/*QueryRewrite':STRAIGHT_JOIN*/1)"
        with CaptureLastQuery() as cap, connection.cursor() as cursor:
            cursor.execute(query)
        cap.query == query

    def test_can_monkey_patch_is_idempotent(self):
        patch_CursorWrapper_execute()

        with CaptureLastQuery() as cap, connection.cursor() as cursor:
            cursor.execute(
                "SELECT 1 FROM DUAL WHERE (/*QueryRewrite':label=hi*/1)")
        assert cap.query == "SELECT /*hi*/ 1 FROM DUAL WHERE (1)"
